
/*
 * ldap gconf backend
 *
 * Copyright (C) 2000 Ben Martin (aka monkeyiq)
 *
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */
 
// glib
#include <glib.h>

// c
#include <stdio.h>
#include <time.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <limits.h>

// local
#include "ldap-backend.h"
#include "ldap_gconf_support.h"
#include "ldap_gconf_ldapmod.h"

/*
 * VTable functions
 */

static void          ldap_gconf_shutdown        (GError** err);

static GConfSource*  ldap_gconf_resolve_address (const gchar* address,
                                      GError** err);

static void          ldap_gconf_lock            (GConfSource* source,
                                      GError** err);

static void          ldap_gconf_unlock          (GConfSource* source,
                                      GError** err);

static gboolean     ldap_gconf_readable         (GConfSource* source,
                                      const gchar* key,
                                      GError** err);

static gboolean     ldap_gconf_writable        (GConfSource* source,
                                      const gchar* key,
                                      GError** err);

static GConfValue*   ldap_gconf_query_value     (GConfSource* source,
                                      const gchar* key,
                                      const gchar** locales,
                                      gchar** schema_name,
                                      GError** err);

static GConfMetaInfo* ldap_gconf_query_metainfo  (GConfSource* source,
                                      const gchar* key,
                                      GError** err);

static void          ldap_gconf_set_value       (GConfSource* source,
                                      const gchar* key,
                                      GConfValue* value,
                                      GError** err);

static GSList*       ldap_gconf_all_entries    (GConfSource* source,
                                     const gchar* dir,
                                     const gchar** locales,
                                     GError** err);

static GSList*       ldap_gconf_all_subdirs     (GConfSource* source,
                                      const gchar* dir,
                                      GError** err);

static void          ldap_gconf_unset_value     (GConfSource* source,
                                      const gchar* key,
                                      const gchar* locale,
                                      GError** err);

static gboolean      ldap_gconf_dir_exists      (GConfSource *source,
                                      const gchar *dir,
                                      GError** err);

static void          ldap_gconf_remove_dir      (GConfSource* source,
                                      const gchar* dir,
                                      GError** err);

static void          ldap_gconf_set_schema      (GConfSource* source,
                                      const gchar* key,
                                      const gchar* schema_key,
                                      GError** err);

static gboolean      ldap_gconf_sync_all        (GConfSource* source,
                                      GError** err);

static void          ldap_gconf_destroy_source  (GConfSource* source);

static void          ldap_gconf_clear_cache     (GConfSource* source);



// all of these need some checking for errors esp LDAP ones.
static GConfBackendVTable ldap_gconf_vtable = {
  ldap_gconf_shutdown,				// ok
  ldap_gconf_resolve_address,		// only simple bind (needs to read URL,username etc from address)
  ldap_gconf_lock,					// ok -- (are there any extended ops for this??
  ldap_gconf_unlock,				// ok
  ldap_gconf_readable,				// not done (always 1)
  ldap_gconf_writable,				// not done (always 1)
  ldap_gconf_query_value,			// yes
  ldap_gconf_query_metainfo,		// yes
  ldap_gconf_set_value,				// yes
	
  ldap_gconf_all_entries,			// ok
  ldap_gconf_all_subdirs,			// ok
	
  ldap_gconf_unset_value,			// ok
  ldap_gconf_dir_exists,			// ok
  ldap_gconf_remove_dir,			// ok
	
  ldap_gconf_set_schema,			// ok
  ldap_gconf_sync_all,				// ok
  ldap_gconf_destroy_source,		// ok
  ldap_gconf_clear_cache			// ok
};


/**************************************************************************/
/**************************************************************************/
/**************************************************************************/
/**************************************************************************/
/**************************************************************************/
/**************************************************************************/
/**************************************************************************/
/**************************************************************************/
static void          
ldap_gconf_shutdown (GError** err)
{
  gconf_log(GCL_INFO, _("Unloading ldap backend module."));
}

/**************************************************************************/
/**************************************************************************/
/**************************************************************************/
/**************************************************************************/
static void
ldap_gconf_lock (GConfSource* source,
      GError** err)
{
	gconf_log(	GCL_DEBUG, "ldap_gconf_lock _top ");
}

/**************************************************************************/
/**************************************************************************/
/**************************************************************************/
/**************************************************************************/
static void
ldap_gconf_unlock (GConfSource* source,
        GError** err)
{
		gconf_log(	GCL_DEBUG, "ldap_gconf_unlock _top ");

}

/**************************************************************************/
/**************************************************************************/
/**************************************************************************/
/**************************************************************************/
static gboolean
ldap_gconf_readable (GConfSource* source,
          const gchar* key,
          GError** err)
{
	gconf_log(	GCL_DEBUG, "ldap_gconf_readable _top ");

	/* FIXME */
  	return TRUE;
}

/**************************************************************************/
/**************************************************************************/
/**************************************************************************/
/**************************************************************************/
static gboolean
ldap_gconf_writable (GConfSource* source,
           const gchar* key,
           GError** err)
{
	gconf_log(	GCL_DEBUG, "ldap_gconf_writable _top ");

	/* FIXME */
	return TRUE;
}

#include <popt.h>
enum popt_option_hit {
	_POPT_USERNAME = 1,
	_POPT_PASSWORD,
	_POPT_SECONDS_TO_CACHE,
	_POPT_MAX_DEPTH,
	_POPT_KEY_BASE,
	_POPT_SERIALIZE_STRING_MARSHAL,
	_POPT_SERIALIZE_INT_MARSHAL,
	_POPT_SERIALIZE_BOOL_MARSHAL,
	_POPT_SERIALIZE_FLOAT_MARSHAL,
	_POPT_SERIALIZE_PAIR_MARSHAL,
	_POPT_SERIALIZE_LIST_MARSHAL,
	_POPT_SERIALIZE_ALL_MARSHALED,
	_POPT_END_LIST
};

static struct poptOption ldap_poptOptions[] = {

	/** ******************************* */
	
	{
		"login",'l',
		POPT_ARG_STRING,NULL,_POPT_USERNAME,NULL,NULL,
	},
	{
		"password",'p',
		POPT_ARG_STRING,	NULL,_POPT_PASSWORD,	NULL, NULL,
	},
	{
		"seconds-to-cache",'s',
		POPT_ARG_INT,	NULL,0,	NULL, NULL,
	},
	{
		"max-depth",'d',
		POPT_ARG_INT,	NULL,0,	NULL, NULL,
	},
	{
		"key-base",'k',
		POPT_ARG_STRING,	NULL,_POPT_KEY_BASE,	NULL, NULL,
	},

	/** *******************/

	{
		"marshal-string", 0,
		POPT_ARG_NONE,	NULL,_POPT_SERIALIZE_STRING_MARSHAL,	NULL, NULL,
	},
	{
		"marshal-int", 0,
		POPT_ARG_NONE,	NULL,_POPT_SERIALIZE_INT_MARSHAL,	NULL, NULL,
	},
	{
		"marshal-bool", 0,
		POPT_ARG_NONE,	NULL,_POPT_SERIALIZE_BOOL_MARSHAL,	NULL, NULL,
	},
	{
		"marshal-float", 0,
		POPT_ARG_NONE,	NULL,_POPT_SERIALIZE_FLOAT_MARSHAL,	NULL, NULL,
	},
	{
		"marshal-pair", 0,
		POPT_ARG_NONE,	NULL,_POPT_SERIALIZE_PAIR_MARSHAL,	NULL, NULL,
	},
	{
		"marshal-list", 0,
		POPT_ARG_NONE,	NULL,_POPT_SERIALIZE_LIST_MARSHAL,	NULL, NULL,
	},
	{
		"marshal-all", 0,
		POPT_ARG_NONE,	NULL,_POPT_SERIALIZE_ALL_MARSHALED,	NULL, NULL,
	},
	
	/** *******************/
	
	{
		NULL, 0,	0, NULL,	0, NULL, NULL,
	}
};


/**
 * Parse backend specific options.
 *
 * Returns 1 on error, 0 if ok
*/
int 
ldap_gconf_parse_args( LDAPSource *ld_source, GError** err, const gchar* address)
{
	char* backend_path 	= NULL;
	int argc 			= 0;
	char** argv			= NULL;
	int popt_err		= 0;
	int ret 			= 0;
	poptContext popt_context = 0;
	
	gconf_log(	GCL_DEBUG, "ldap_gconf_parse_args called");

	
	/* Get the backend specific part */
	backend_path = gconf_address_resource(address);
  	if (backend_path == NULL || !strlen(backend_path))
    {
      gconf_set_error(err, GCONF_ERROR_BAD_ADDRESS, 
		_("You must supply a location for the LDAP server in the address `%s'"), address);
      return 1;
    }
	
	/* 
	 * To stop those nagging pains, popt expects argv format, 
	 * so make sure there is a simulated application name
	 */
	backend_path = g_strchug( backend_path );
	if( backend_path[0] == '-' ) {
		gchar *tmp = g_strconcat("dummyAppName  ", backend_path, NULL);
		g_free( backend_path );
		backend_path = tmp;
	}

	/* defaults */
	ld_source->serialize_as = SERIALIZE_ALL_NO_MARSHAL;
	ld_source->seconds_to_cache = 60*5;
	ld_source->max_levels_of_depth = 100;
	
	
	/* Setup some the be automatic */
	ldap_poptOptions[_POPT_SECONDS_TO_CACHE-1].arg 		= &ld_source->seconds_to_cache;
	ldap_poptOptions[_POPT_MAX_DEPTH-1].arg 			= &ld_source->max_levels_of_depth;

	if((popt_err = poptParseArgvString( backend_path, &argc, &argv )) >= 0) 
	{
		gconf_log(	GCL_DEBUG, "ldap_gconf_parse_args argc: %d",argc);
		popt_context =  poptGetContext( "ldap-gconf-backend", argc, argv, ldap_poptOptions, 0 );

		while((popt_err = poptGetNextOpt(popt_context)) > 0) 
		{
			const char *ca = NULL;

			#define LDAP_GCONF_PARSE_ARGS_REQUIRED_PARAM_MISSING(ARG) { \
				ret = 1; \
				gconf_log(	GCL_DEBUG, "ldap_gconf_parse_args required arg missing %s\n", \
					ldap_poptOptions[ARG-1].longName); \
				gconf_set_error(err, GCONF_ERROR_PARSE_ERROR,  \
					_("A required argument is missing for %s\n"), ldap_poptOptions[ARG-1].longName); \
				}

			#define LDAP_GCONF_PARSE_ARGS_REQUIRED_HANDLE( STRDST, CASE ) \
					if((ca = poptGetOptArg(popt_context))) { \
						STRDST = g_strdup(ca); \
					} \
					else \
						LDAP_GCONF_PARSE_ARGS_REQUIRED_PARAM_MISSING(CASE); \
					break;					
			
			#define LDAP_GCONF_PARSE_ARGS_REQUIRED_UNSET_BIT( BZZ ) \
				if ((( ld_source->serialize_as & BZZ) == BZZ)) { \
					ld_source->serialize_as ^= BZZ; \
				} \
				break;
				
			switch( popt_err ) 
			{
				
				case _POPT_USERNAME: 
					LDAP_GCONF_PARSE_ARGS_REQUIRED_HANDLE(ld_source->username,_POPT_USERNAME)
				case _POPT_PASSWORD:
					LDAP_GCONF_PARSE_ARGS_REQUIRED_HANDLE(ld_source->password,_POPT_PASSWORD)
				case _POPT_KEY_BASE:
					LDAP_GCONF_PARSE_ARGS_REQUIRED_HANDLE(ld_source->key_base,_POPT_KEY_BASE)
				case _POPT_SERIALIZE_STRING_MARSHAL:
					LDAP_GCONF_PARSE_ARGS_REQUIRED_UNSET_BIT(SERIALIZE_STRING_NO_MARSHAL);
				case _POPT_SERIALIZE_INT_MARSHAL:
					LDAP_GCONF_PARSE_ARGS_REQUIRED_UNSET_BIT(SERIALIZE_INT_NO_MARSHAL);
				case _POPT_SERIALIZE_BOOL_MARSHAL:
					LDAP_GCONF_PARSE_ARGS_REQUIRED_UNSET_BIT(SERIALIZE_BOOL_NO_MARSHAL);
				case _POPT_SERIALIZE_FLOAT_MARSHAL:
					LDAP_GCONF_PARSE_ARGS_REQUIRED_UNSET_BIT(SERIALIZE_FLOAT_NO_MARSHAL);
				case _POPT_SERIALIZE_PAIR_MARSHAL:
					LDAP_GCONF_PARSE_ARGS_REQUIRED_UNSET_BIT(SERIALIZE_PAIR_NO_MARSHAL);
				case _POPT_SERIALIZE_LIST_MARSHAL:
					LDAP_GCONF_PARSE_ARGS_REQUIRED_UNSET_BIT(SERIALIZE_LIST_NO_MARSHAL);
				case _POPT_SERIALIZE_ALL_MARSHALED:
					ld_source->serialize_as = SERIALIZE_ALL_AS_MARSHALED;
					break;
			}

			
		} // while poptGetNextOpt 
	} // if poptParseArgvString 
					

	if( popt_err < -1 ) {
		gconf_log(	GCL_DEBUG, "ldap_gconf_parse_args popt_err %s: %s\n",
               poptBadOption(popt_context, POPT_BADOPTION_NOALIAS),
               poptStrerror(popt_err));
		gconf_set_error(err, GCONF_ERROR_PARSE_ERROR, 
			_("Invalid backend args %s: %s\n"),
               poptBadOption(popt_context, POPT_BADOPTION_NOALIAS),
               poptStrerror(popt_err));
		ret = 1;		
	}
	
	/* And now we get the remains to use for connecting to servers as URLs server:port */
	if(!ret)
	{
		const char ** leftovers = poptGetArgs(popt_context);
		
		if( !leftovers || !leftovers[0]) {
			gconf_log(	GCL_DEBUG, "ldap_gconf_parse_args no server provided in %s\n",backend_path);
			gconf_set_error(err, GCONF_ERROR_PARSE_ERROR, 
				_("No server location provided in args: %s\n"),	backend_path);
			ret = 1;		
		}
		
		if( !ret ) {
			/* FIXME: hmm check to see if g_strjoinv 2nd arg is really const */ 
			ld_source->server_url = g_strjoinv(" ", (gchar **)leftovers); 
		}
	}
	
	/* A final sanity check */
	if(!ret) {
		if( 
				!ld_source->server_url
			||	!ld_source->username
			||	!ld_source->password
			||	!ld_source->key_base)
		{
			gconf_log(	GCL_DEBUG, "ldap_gconf_parse_args not enough args in %s\n",backend_path);
			gconf_set_error(err, GCONF_ERROR_PARSE_ERROR, 
				_("not enough args in args: %s\n"),	backend_path);
			ret = 1;		
		}
	}
	
	gconf_log(	GCL_DEBUG, "ldap_gconf_parse_args done");
	g_free(backend_path);
	if( argv ) free( argv );
	poptFreeContext( popt_context );
	
	return ret;
}


/**************************************************************************/
/**************************************************************************/
/**************************************************************************/
/**************************************************************************/
static GConfSource*  
ldap_gconf_resolve_address (const gchar* address, GError** err)
{
	LDAPSource *ld_source = NULL;
	GConfSource *source = NULL;
	int ldap_err = 0;
	
	gconf_log(	GCL_DEBUG, "ldap_gconf_resolve_address called");
	
	/* Create an instance */	
	ld_source = (LDAPSource *) g_new0 (LDAPSource,1);
	source = (GConfSource *) ld_source;



	ld_source->serialize_as = SERIALIZE_ALL_NO_MARSHAL;
	if(ldap_gconf_parse_args( ld_source, err, address ))
	{
		g_free( ld_source );
		return NULL;
	}

	gconf_log(	GCL_DEBUG, "ldap_gconf_resolve_address server_url:%s ",ld_source->server_url);
	gconf_log(	GCL_DEBUG, "ldap_gconf_resolve_address username:%s ",ld_source->username);
	gconf_log(	GCL_DEBUG, "ldap_gconf_resolve_address password:%s ",ld_source->password);
	gconf_log(	GCL_DEBUG, "ldap_gconf_resolve_address key_base:%s ",ld_source->key_base);
	gconf_log(	GCL_DEBUG, "ldap_gconf_resolve_address seconds_to_cache:%d ",ld_source->seconds_to_cache);
	gconf_log(	GCL_DEBUG, "ldap_gconf_resolve_address max_levels_of_depth:%d ",ld_source->max_levels_of_depth);
	

	/* open a connection  */
	if ((ld_source->ld = ldap_init(ld_source->server_url, LDAP_PORT)) == NULL)
	{
		gconf_set_error(err, GCONF_ERROR_NO_SERVER, 
			 _("Can not open connection to LDAP server at address `%s'. Error was `%s'"), 
			address, strerror(errno));
		return NULL;
    }
	
	gconf_log(	GCL_DEBUG, "ldap_gconf_resolve_address 5. ");
	gconf_log(	GCL_DEBUG, "ldap_gconf_resolve_address 5. %s ",ld_source->username);
	gconf_log(	GCL_DEBUG, "ldap_gconf_resolve_address 5. %s ",ld_source->password);

	/* setup client side cache */
	if( -1 == ldap_enable_cache(ld_source->ld, ld_source->seconds_to_cache, 0)) {
		gconf_log(	GCL_ERR, "ldap_gconf_resolve_address Can not setup client side cache, expect slow preformance.");
	}

	/* authenticate using "Simple" auth method */
	if ((ldap_err = ldap_simple_bind_s(ld_source->ld, ld_source->username, ld_source->password))
		!= LDAP_SUCCESS) 
	{
		gconf_log(	GCL_WARNING, "ldap_gconf_resolve_address Can not authenticate to LDAP");
		_ldap_gconf_set_error(err, GCONF_ERROR_BAD_ADDRESS, ldap_err, 
			_("Can not open connection to LDAP server at address `%s' using simple auth binding."),
			address);
		return NULL;
	}
	
	gconf_log (GCL_DEBUG, _("Opened LDAP source at root %s"), address);
	return source;
}


/**************************************************************************/
/**************************************************************************/
/**************************************************************************/
/**************************************************************************/
// perform an LDAP search for this value. //
//
static GConfValue* 
ldap_gconf_query_value (GConfSource* source,
             const gchar* key,
             const gchar** locales,
             gchar** schema_name,
             GError** err)
{
	MAKE_INSTANCE_VARS( source, NULL )
	LDAPMessage     *res;
	GConfValue* retval = NULL;
	gchar* ldap_key = NULL;

	gconf_log(	GCL_DEBUG, "ldap_gconf_query_value called ");
	ldap_key = _key_to_ldap_dn( ld_source, err, key );
	gconf_log(	GCL_DEBUG, "ldap_gconf_query_value key:%s ",ldap_key);
	
	if(_ldap_gconf_lookup_key( ld_source, err, 0, NULL, NULL, ldap_key, &res )) {
		g_free(ldap_key);
		return NULL;
	}
	gconf_log(	GCL_DEBUG, "ldap_gconf_query_value key:%s was looked up ",ldap_key);
	
	retval = _ldap_gconf_get_GConf_Value( ld_source, err, res, &retval);
	gconf_log(	GCL_DEBUG, "ldap_gconf_query_value 2");
	gconf_log(	GCL_DEBUG, "ldap_gconf_query_value key:%s was made into retval %s",
		ldap_key,
		retval ? gconf_value_to_string(retval) : "retval==NULL"
	);
	
	/* free memory allocated for search results */
	ldap_msgfree(res);	
	g_free(ldap_key);
	
	return retval;
}


/**************************************************************************/
/**************************************************************************/
/**************************************************************************/
/**************************************************************************/
static GConfMetaInfo*
ldap_gconf_query_metainfo  (GConfSource* source, const gchar* key,
                 GError** err)
{
	MAKE_INSTANCE_VARS( source, NULL )
	LDAPMessage     *res = NULL;
	LDAPMessage     *entry;
	gchar* ldap_key 	= NULL;
	GConfMetaInfo* ret 	= NULL;
	gint error 			= 0;
	gchar* schema_name	= NULL;
	gchar* mod_user		= NULL;
	GTime mod_time 		= -1;
	gchar* mod_time_str	= NULL;
	int ldap_err		= -1;
	int numfound 		= 0;
	char *attrs[] = { 
		LDAP_GCONF_META_CREATORNAME_KEY, 
		LDAP_GCONF_META_MODIFIERSNAME_KEY, 
		LDAP_GCONF_META_MODIFYTIMESTAMP, 
		LDAP_GCONF_META_SCHEMA_KEY,
		NULL 
	};

	gconf_log(	GCL_DEBUG, "ldap_gconf_query_metainfo called ");
	ldap_key = _key_to_ldap_dn( ld_source, err, key );
	gconf_log(	GCL_DEBUG, "ldap_gconf_query_metainfo key:%s ",ldap_key);

	if(_ldap_gconf_lookup_key( ld_source, err, LDAP_SCOPE_BASE,
			"(objectclass=externalobject)", attrs, ldap_key, &res)) {
		g_free( ldap_key );
		return NULL;
	}
	
	gconf_log(	GCL_DEBUG, "ldap_gconf_query_metainfo key:%s was looked up ",ldap_key);

	
	/* did we get anything ? */
	numfound = ldap_count_entries( ld_source->ld, res ); 
	if ( 1 != numfound ) {
		gchar *dn = ldap_get_dn(ld, res);
		gconf_set_error(err, GCONF_ERROR_BAD_KEY, 
			_("Failed to find correct data for key `%s' expected one, got %d"), 
			dn, numfound );
		g_free(dn);
		error = 1;
	}

	if(!error) {
	
		/* There is only one entry */
		entry = ldap_first_entry( ld, res );
		
		/* Grab it from result set */
		if((mod_user = _ldap_gconf_get_value( ld, err, entry, LDAP_GCONF_META_MODIFIERSNAME_KEY ))) {
			if((schema_name = _ldap_gconf_get_value( ld, err, entry, LDAP_GCONF_META_SCHEMA_KEY ))) {
				if((mod_time_str = _ldap_gconf_get_value( ld, err, entry, LDAP_GCONF_META_MODIFYTIMESTAMP ))) {
					mod_time = _convert_from_ldap_time( mod_time_str);
				}
			}
		}
		
		/* Did we get it all ? */
		if( mod_user && schema_name && mod_time_str && mod_time!=-1) {
			gconf_log(	GCL_DEBUG, "ldap_gconf_query_metainfo got it all ");
			ret = gconf_meta_info_new();
			gconf_meta_info_set_schema( ret, schema_name );
			gconf_meta_info_set_mod_user( ret, mod_user );
			gconf_meta_info_set_mod_time( ret, mod_time);
		}
		else {
			error = 1;
		}
	}
	
	/* free memory allocated for search results */
	if(res) 			ldap_msgfree(res);	
	if(mod_user) 		g_free(mod_user);
	if(schema_name)		g_free(schema_name);
	if(mod_time_str)	g_free(mod_time_str);
	if(ldap_key)		g_free(ldap_key);

	return error ? NULL : ret;
}

/**************************************************************************/
/**************************************************************************/
/**************************************************************************/
/**************************************************************************/
static void          
ldap_gconf_set_value (GConfSource* source, const gchar* key, GConfValue* value,
           GError** err)
{
	MAKE_INSTANCE_VARS( source, NULL )
	gchar* ldap_key = NULL;
	static serialize_from_gconf_value_vtable vtab = {
		SERIALIZE_FROM_GCONF_VALUE_VTABLE_VERSION_1,
		_ldap_gconf_type_and_val_f,
		_ldap_gconf_pair_f,
		_ldap_gconf_list_f, 
		_ldap_gconf_marshaled_f
	};
	
	gconf_log(	GCL_DEBUG, "ldap_gconf_set_value vtable. called ");
	ldap_key = _key_to_ldap_dn( ld_source, err, key );
	gconf_log(	GCL_DEBUG, "ldap_gconf_set_value vtable. 1 ");

	serialize_from_gconf_value( 
		err,
		ldap_key,
		value,
		ld_source->serialize_as,
		ld_source,
		&vtab);
	
	gconf_log(	GCL_DEBUG, "ldap_gconf_set_value vtable. 2 ");
	g_free(ldap_key);
}



/**************************************************************************/
/**************************************************************************/
/**************************************************************************/
/**************************************************************************/
static GSList*             
ldap_gconf_all_entries    (GConfSource* source,
                const gchar* key,
                const gchar** locales,
                GError** err)
{
	MAKE_INSTANCE_VARS( source, NULL )
	LDAPMessage     *res = NULL;
	LDAPMessage     *e;
	GSList* ret 	= NULL;
	gint error		= 0;
	gchar* ldap_key = NULL;
	int numfound 	= 0;

	gconf_log(	GCL_DEBUG, "ldap_gconf_all_entries called ");
	ldap_key = _key_to_ldap_dn( ld_source, err, key );
	gconf_log(	GCL_DEBUG, "ldap_gconf_all_entries key:%s ",ldap_key);

	if(_ldap_gconf_lookup_key( ld_source, err, LDAP_SCOPE_ONELEVEL, 
		"(&(objectclass=externalobject)(type=*))", NULL, ldap_key, &res )) {
		g_free(ldap_key);
		return NULL;
	}
	g_free(ldap_key);

	/* did we get anything ? */
	if ((numfound = ldap_count_entries( ld_source->ld, res )) == -1) {
		gconf_log(	GCL_DEBUG, "ldap_gconf_all_entries didn't get anything");
		gconf_set_error(err, GCONF_ERROR_BAD_KEY, 
			_("Failed to find any data for key `%s'"), ldap_get_dn(ld, res));
		error = 1;
	}
	gconf_log(	GCL_DEBUG, "ldap_gconf_all_entries numfound:%d",numfound);

	if( !error ) {
	
		/* step through each entry returned */
		for ( e = ldap_first_entry( ld, res ); e != NULL;
			e = ldap_next_entry( ld, e ) ) 
		{
			GConfValue* v;
			gchar* dn;
			
			dn = ldap_get_dn( ld, e );
			gconf_log(	GCL_DEBUG, "ldap_gconf_all_entries  got dn:%s",dn);
			free( dn );
	
			v = ldap_gconf_reify_gvalue( ld_source, err, e );
			if(!v) {
				gpointer gptr;
				
				gconf_log(	GCL_DEBUG, "ldap_gconf_all_entries, falied to reify a GConvValue");

				/* Delete this list that has been created so far */
				while(( gptr = g_slist_nth( ret ,0 ))) {
					ret = g_slist_remove( ret, gptr );
					gconf_value_free((GConfValue*)gptr);
				}
				g_slist_free( ret );
				ret = NULL;
				error = 1;
				break;
			}
			ret = g_slist_append(ret, v);
		}		
	}
	
	/* free the search results */
	if(res) 	ldap_msgfree( res );
		
	return error ? NULL : ret;
}


/**************************************************************************/
/**************************************************************************/
/**************************************************************************/
/**************************************************************************/
static GSList*
ldap_gconf_all_subdirs     (GConfSource* source,
                 const gchar* key,
                 GError** err)
{
	MAKE_INSTANCE_VARS( source, NULL )
	LDAPMessage* res 	= NULL;
	GSList* 	 ret 	= NULL;
	gint error			= 0;
	LDAPMessage* e		= NULL;
	gchar* 		ldap_key= NULL;
	int numfound 		= 0;

	gconf_log(	GCL_DEBUG, "ldap_gconf_all_subdirs _top ");
	ldap_key = _key_to_ldap_dn( ld_source, err, key );
	gconf_log(	GCL_DEBUG, "ldap_gconf_all_subdirs key:%s ",ldap_key);

	if(_ldap_gconf_lookup_key( ld_source, err, LDAP_SCOPE_ONELEVEL, 
		"(objectclass=externalobject)", NULL, ldap_key, &res )) {
		g_free(ldap_key);
		return NULL;
	}
	g_free(ldap_key);
	ldap_key = NULL;
	

	/* did we get anything ? */
	if ((numfound = ldap_count_entries( ld_source->ld, res )) == -1) {
		gconf_log(	GCL_DEBUG, "ldap_gconf_all_subdirs didn't get anything");
		gconf_set_error(err, GCONF_ERROR_BAD_KEY, 
			_("Failed to find any data for key `%s'"), ldap_get_dn(ld, res));
		return NULL;
	}
	gconf_log(	GCL_DEBUG, "ldap_gconf_all_subdirs numfound:%d",numfound);

    /* step through each entry returned */
	for ( e = ldap_first_entry( ld, res ); e != NULL;
		e = ldap_next_entry( ld, e ) ) 
	{
		gchar* dn;
		char** sv;

		sv = ldap_get_values( ld, e, key );
		if( 0 != ldap_count_values(sv)) {
			gconf_log(	GCL_DEBUG, "ldap_gconf_all_subdirs there is a type for attrib");
			ldap_value_free(sv);
			continue;
		}
		ldap_value_free(sv);
		sv = NULL;

		dn = ldap_get_dn( ld, e );
		gconf_log(	GCL_DEBUG, "ldap_gconf_all_entries  got dn of==> %s",dn);
		sv = ldap_explode_dn( dn, 1 );
		if( !sv || !sv[0] ) {
			gpointer gptr;
			gconf_log(	GCL_DEBUG, "ldap_gconf_all_entries  !sv");
			gconf_set_error(err, GCONF_ERROR_CORRUPT, 
				_("rdn could not be acquired for key `%s'"), dn);
			ldap_value_free(sv);
			free( dn );

			/* Delete this list that has been created so far */
			while(( gptr = g_slist_nth( ret ,0 ))) {
				ret = g_slist_remove( ret, gptr );
				g_free((gchar*)gptr);
			}
			g_slist_free( ret );
			ret = NULL;
			error = 1;
			break;
		}
		ret = g_slist_append(ret, g_strdup(sv[0]));
		ldap_value_free(sv);
		free( dn );
	}		
	
	/* free the search results */
	if(res) ldap_msgfree( res );
		
	return error ? NULL : ret;
}


/**************************************************************************/
/**************************************************************************/
/**************************************************************************/
/**************************************************************************/
static void          
ldap_gconf_unset_value     (GConfSource* source,
                 const gchar* key,
                 const gchar* locale,
                 GError** err)
{
	MAKE_INSTANCE_VARS( source, NULL )
	gchar* ldap_key 	= NULL;
	int ldap_err 		= 0;

	gconf_log(	GCL_DEBUG, "ldap_gconf_unset_value called ");
	ldap_key = _key_to_ldap_dn( ld_source, err, key );
	gconf_log(	GCL_DEBUG, "ldap_gconf_unset_value key:%s ",ldap_key);

	if((ldap_err = ldap_delete_s( ld, ldap_key )) != LDAP_SUCCESS) {
		gconf_log(	GCL_DEBUG, "ldap_gconf_unset_value _failed");
		_ldap_gconf_set_error( err, GCONF_ERROR_FAILED, ldap_err, "",NULL);
	}

	g_free(ldap_key);
	gconf_log(	GCL_DEBUG, "ldap_gconf_unset_value done");
	return;
}

/**************************************************************************/
/**************************************************************************/
/**************************************************************************/
/**************************************************************************/
static gboolean
ldap_gconf_dir_exists      (GConfSource*source,
                 const gchar* key,
                 GError** err)
{
	MAKE_INSTANCE_VARS( source, NULL )
	gboolean 	 ret 	= 0;
	LDAPMessage *res	= NULL;
	gchar* ldap_key 	= NULL;
	int ldap_err 		= 0;

	gconf_log(	GCL_DEBUG, "ldap_gconf_dir_exists _top ");
	ldap_key = _key_to_ldap_dn( ld_source, err, key );
	gconf_log(	GCL_DEBUG, "ldap_gconf_dir_exists key:%s ",ldap_key);

	/*
	 * search the database, ldap_search_s() is used directly because we have 
	 * the inverse semantics than usual, we actually want the lookup to fail 
	*/
	if ( (ldap_err = ldap_search_s( 
		ld, 
		ldap_key, 
		LDAP_SCOPE_BASE,
		"(objectclass=externalobject)", 
		NULL,
    	0, 
		&res))
	!= LDAP_SUCCESS) 
	{
		if(ldap_err != LDAP_NO_SUCH_OBJECT) {
			gconf_log(	GCL_DEBUG, "ldap_gconf_remove_dir search error ");
			_ldap_gconf_set_error( err, GCONF_ERROR_BAD_KEY, ldap_err, "",NULL);
		}
		ret = 0;
	}

	/* did we get anything ? */
	if (1 == ldap_count_entries( ld_source->ld, res )) {
		ret = 1;
	}
	
	/* free the search results */
	if(res) 		ldap_msgfree( res );
	if(ldap_key) 	g_free(ldap_key);
		
	return ret;
}

int
ldap_gconf_remove_dir_cmp_f(char* a, char* b) 
{
	gconf_log(	GCL_DEBUG, "ldap_gconf_remove_dir_cmp_f _top ");
	gconf_log(	GCL_DEBUG, "ldap_gconf_remove_dir_cmp_f a:%s ",a);
	gconf_log(	GCL_DEBUG, "ldap_gconf_remove_dir_cmp_f b:%s ",b);
	return g_strcasecmp(a,b);

}




/**************************************************************************/
/**************************************************************************/
/**************************************************************************/
/**************************************************************************/
static void          
ldap_gconf_remove_dir      (GConfSource* source,
                 const gchar* key,
                 GError** err)
{
	MAKE_INSTANCE_VARS( source, NULL )
	LDAPMessage* res 	= NULL;
	gint error			= 0;
	gchar* ldap_key 	= NULL;
	int numfound 		= 0;
	int ldap_err 		= 0;
	char *attrs[] = { "dn", NULL };
	

	gconf_log(	GCL_DEBUG, "ldap_gconf_remove_dir called ");
	ldap_key = _key_to_ldap_dn( ld_source, err, key );
	gconf_log(	GCL_DEBUG, "ldap_gconf_remove_dir key:%s ",ldap_key);

	/*
	 * search the database, ldap_search_s() is used directly because we have 
	 * the inverse semantics than usual, we actually want the lookup to fail 
	*/
	if ( (ldap_err = ldap_search_s( 
		ld, 
		ldap_key, 
		LDAP_SCOPE_SUBTREE,
		"(objectclass=externalobject)", 
		attrs,
    	1, 
		&res))
	!= LDAP_SUCCESS) 
	{
		if(ldap_err != LDAP_NO_SUCH_OBJECT) {
			gconf_log(	GCL_DEBUG, "ldap_gconf_remove_dir search error ");
			_ldap_gconf_set_error( err, GCONF_ERROR_BAD_KEY, ldap_err, "",NULL);
		}
		error = 1;
	}
	else 
	{
		/* If there is a subtree we must delete it now. */
		if ((numfound= ldap_count_entries( ld_source->ld, res )))
		{
			gconf_log(	GCL_DEBUG, "ldap_gconf_remove_dir recurse work todo numfound:%d",numfound);
			if(_ldap_gonf_delete_recursive( ld_source, err, res )) {
				error = 1;
			}
		}
	}

	if( !error ) {
		/* Delete this key */
		if((ldap_err = ldap_delete_s( ld, ldap_key )) != LDAP_SUCCESS) {
			_ldap_gconf_set_error( err, GCONF_ERROR_BAD_KEY, ldap_err, "",NULL);
		}
	}
	
	/* free the search results */
	if(res) 		ldap_msgfree( res );
	if(ldap_key) 	g_free(ldap_key);
		
}

/**************************************************************************/
/**************************************************************************/
/**************************************************************************/
/**************************************************************************/
static void          
ldap_gconf_set_schema      (GConfSource* source,
                 const gchar* key,
                 const gchar* schema_key,
                 GError** err)
{
	MAKE_INSTANCE_VARS( source, NULL )
	GPtrArray* a = NULL;

	gchar* meta_key = _ldap_gconf_make_metakey(key);
	if(meta_key) {
		gchar* ldap_key = NULL;
		
		gconf_log(	GCL_DEBUG, "ldap_gconf_set_schema called ");
		ldap_key = _key_to_ldap_dn( ld_source, err, key );
		gconf_log(	GCL_DEBUG, "ldap_gconf_set_schema ldap_key:%s ",ldap_key);

		if( ldap_key ) {
			a=ldap_gconf_ldapmodvec_new();
			ldap_gconf_ldapmodvec_addmod(a, LDAP_MOD_REPLACE, LDAP_GCONF_META_SCHEMA_KEY, meta_key, NULL);
			ldap_gconf_ldapmodvec_addmod(a, LDAP_MOD_REPLACE, LDAP_GCONF_OBJECTCLASS_KEY, LDAP_GCONF_OBJECTCLASS, NULL);
			_ldap_gconf_execute_mods( ld_source, err, ldap_key, a );	
			g_free(ldap_key);
			ldap_gconf_ldapmodvec_free(a);
		}
		g_free(meta_key);
	}
	else 
	{
		_ldap_err_no_ram( err );
	}
}



/**************************************************************************/
/**************************************************************************/
/**************************************************************************/
/**************************************************************************/
static gboolean      
ldap_gconf_sync_all        (GConfSource* source,
                 GError** err)
{
	LDAPSource* ld_source = (LDAPSource*)source;
	gconf_log(	GCL_DEBUG, "ldap_gconf_sync_all called ");
	if(ld_source) {
		ldap_flush_cache( ld_source->ld );
	}
	return 1;
}


/**************************************************************************/
/**************************************************************************/
/**************************************************************************/
/**************************************************************************/
static void          
ldap_gconf_destroy_source  (GConfSource* source)
{
	LDAPSource* ld_source = (LDAPSource*)source;
	int ldap_err = 0;

	g_return_if_fail(source != NULL);
	gconf_log(	GCL_DEBUG, "ldap_gconf_destroy_source called ");

	if (((ldap_err = ldap_unbind_s(ld_source->ld)))	!= LDAP_SUCCESS) 
	{
		gconf_log(	GCL_WARNING, "ldap_gconf_destroy_source failed to unbind! Addr: %s",ld_source->server_url);
/*		_ldap_gconf_set_error(err, GCONF_ERROR_BAD_ADDRESS, ldap_err, 
			_("Can not unbind LDAP address `%s'."),
			ld_source->server_url);
*/		
	}
	
	g_free(ld_source->server_url);
	g_free(ld_source->username);
	g_free(ld_source->password);
	g_free(ld_source->key_base);
	g_free(ld_source);
}

/**************************************************************************/
/**************************************************************************/
/**************************************************************************/
/**************************************************************************/
static void
ldap_gconf_clear_cache     (GConfSource* source)
{
	LDAPSource* ld_source = (LDAPSource*)source;
	gconf_log(	GCL_DEBUG, "ldap_gconf_clear_cache called ");
	
	if(ld_source) {
		ldap_flush_cache(ld_source->ld);
	}
}

/**************************************************************************/
/**************************************************************************/
/**************************************************************************/
/**************************************************************************/
/* Initializer */

G_MODULE_EXPORT const gchar*
g_module_check_init (GModule *module)
{
  gconf_log(GCL_INFO, _("Initializing ldap backend module"));
  return NULL;
}

G_MODULE_EXPORT GConfBackendVTable* 
gconf_backend_get_vtable(void)
{
  return &ldap_gconf_vtable;
}



/**************************************************************************/
/**************************************************************************/
/**************************************************************************/
/**************************************************************************/

