
/*
 * ldap gconf backend
 *
 * Copyright (C) 2000 Ben Martin (aka monkeyiq)
 *
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/*
	Note that many of these functions start with _ to show that they are support for the
	library and not directly enterable from the gconf engine.

*/

/* local */
#include "ldap_gconf_support.h"
#include "ldap_gconf_ldapmod.h"
#include "gblob.h"

/* some ldap API stuff uses malloc/free (eh) */
#include <stdlib.h>


/**
 * This function tags on LDAP specific error information to the GError that is passed back
 * to the user.
 *
 * ldap_err is the LDAP error code that occured.
 *
 * all other params are the same as gconf_set_error()
 */
void
_ldap_gconf_set_error(GError** err, GConfError en, int ldap_err, const gchar* fmt, ...)
{
	va_list args;
	gchar* newfmt = NULL;
	gchar* newfmt2 = NULL;
  
  	va_start (args, fmt);
	newfmt = g_strdup_vprintf( fmt, args );
	va_end (args);
	newfmt2 = g_strconcat(newfmt,_(" ldap specific error number: `%d' error string `%s'"),NULL);

	gconf_set_error(err, en, newfmt2, ldap_err, ldap_err2string(ldap_err));
	
	g_free(newfmt);
	g_free(newfmt2);
	return;
}

/**
 * Debug code that dumps out a char** to the gconf log fle.
 *
 * vec The vector of strings to log.
 */
void
g_strv_show( gchar** vec ) {
	int i=0;
	
	g_return_if_fail( vec );
	
	for(i=0; vec[i]; i++ ) {
		gconf_log(	GCL_DEBUG, "g_strv_show i:%d vec[i]:%s:",i,vec[i]);
	}
}

/**
 * get the length of a string vector
 *
 * vec The vector of strings
 */
int
g_strv_length( gchar** vec ) {
	int i=0;
	
	g_assert( vec );
	
	for(i=0; vec[i]; i++ ) {
	}
	return i;
}

/**
 * Produce and return a verbatim copy of strv.
 *
 * Caller must g_strfreev() return value.
 */
gchar**
g_strv_copy( gchar** strv ) 
{
	int leng = g_strv_length(strv);
	int i=0;
	
	gchar **ret = g_new0( gchar*, leng+1 );
	for( i=0; i < leng; i++ ) 
	{
		ret[i] = g_strdup( strv[i] );
	}
	
	return ret;
}

/**
 * reverse a gchar**
 *
 * Note that the input vector of strings is moved in place. There is the 
 * expectation that "vec" is NULL terminated.
 *
 */
gchar** g_strv_reverse( gchar ** vec ) {
	int i=0, max=0, half=1;
	gchar *tmp;

	gconf_log(	GCL_DEBUG, "g_strv_reverse called");
	g_strv_show(vec);
	
	for(i=0; vec[i] != NULL; ) {
		++i;
	}
	max = --i;
	half = max/2;
	gconf_log(	GCL_DEBUG, "g_strv_reverse i:%d max:%d half:%d",i,max,half);
	
	for( ; i > half; --i) {
		tmp = vec[max-i];
		vec[max-i] = vec[i];
		vec[i] = tmp;
	}

	gconf_log(	GCL_DEBUG, "g_strv_reverse done");
	g_strv_show(vec);
	
	return vec;
}



/**
 * Construct a LDAP DN for a given GConf path. 
 *
 * If an error occurs "err" is set and NULL returned.
 *
 * ld_source Instance data for backend
 * err Where to report errors to
 * gk The gconf key to translate into a LDAP dn.
 *
 * returned is a string that is the LDAP dn, caller must g_free() return value.
 */
gchar* 
_key_to_ldap_dn( const LDAPSource* const  ld_source, GError** err, const gchar* gk) 
{
	gchar* r;
	gchar** vec;
	gchar* t;
	gint offset =0;
	
	gconf_log(	GCL_DEBUG, "_key_to_ldap_dn _top gk:%s",gk);
	
	if( !gk || !strlen(gk)) {
		return g_strdup(ld_source->key_base);
	}
	
	/* Remove leading slash char */
	if( gk[offset] == '/') {
		offset++;
	}
	if( gk[offset] == '/') {
		gconf_log(	GCL_EMERG, "Backend has detected a key with two leading / chars key:%s",gk);
		gconf_set_error( err, GCONF_ERROR_BAD_KEY, _("Key can not contain more than one leading slash char. key`%s'"),gk);
		return NULL;
	}
		
	vec = g_strsplit( &gk[offset], "/", ld_source->max_levels_of_depth );
	
	vec = g_strv_reverse(vec);
	t=g_strjoinv( ", dn=", vec ); 
	r = g_strdup_printf("dn=%s, %s", t, ld_source->key_base);
	g_free(t);
	g_strfreev (vec);

	gconf_log(	GCL_DEBUG, "_key_to_ldap_dn _done full dn:%s",r);
	return r;
}


/**
 * Perform a ldap_search() on the database, if there is an error, use "err" to report that
 * to the user, and return 1.
 *
 * The return value from this function is 1 if there is an error, in which case the caller 
 * should return NULL to the gconf engine.
 *
 * err The gconf err to put error information into
 * scope is the LDAP scope to search for, if 0 then it is BASE (ie exact match)
 * ldap_key The dn to lookup
 * res Where to place the attributes for this key.
 */
gint
_ldap_gconf_lookup_key( 
	const LDAPSource* const ld_source, 
	GError** err, 
	int scope,
	gchar *filter,
	gchar** attrs,
	gchar* ldap_key, 
	LDAPMessage** res ) 
{
	int ldap_err = 0;
	if( !scope ) scope = LDAP_SCOPE_BASE;
	
	gconf_log(	GCL_DEBUG, "_ldap_gconf_lookup_key ldap_key:%s",ldap_key);
	/* search the database */
	if ( (ldap_err = ldap_search_s( 
		ld_source->ld, 
		ldap_key, 
		scope,
		filter ? filter : "(objectclass=externalobject)", 
		attrs,
    	0, 
		res))
	!= LDAP_SUCCESS) 
	{
		gconf_log(	GCL_DEBUG, "_ldap_gconf_lookup_key err:%d %s",ldap_err,ldap_err2string(ldap_err));
		_ldap_gconf_set_error( err, GCONF_ERROR_BAD_KEY, ldap_err, "",NULL);
		return 1;
	}
	return 0;
}


/**
 * Given a valid "entry" from a search, this function looks up the attribute using
 * "key" and returns the values for that attribute. Note that the function will 
 * set an error (and return NULL) in case of any problems.
 * 
 * The function assumes that if the attribute is not available then an error has occured.
 *
 * ld_source Instance data for backend
 * err Where to report errors to
 * entry Results from some other search
 * key Name of attribute to lookup
 * 
 * Caller must g_strfreev() the return value.
 */
gchar** 
_ldap_gconf_get_values( LDAP* ld, GError** err, LDAPMessage* entry, gchar* key ) 
{
	char  **vals;
	gchar** ret;
	int numfound = 0;

	/* Determine the datatype of the key */
	vals = ldap_get_values( ld, entry, key );
	numfound = ldap_count_values(vals);
	if( numfound < 1 ) {
		gchar *dn = ldap_get_dn(ld, entry);
		gconf_log(	GCL_DEBUG, "_ldap_gconf_get_values Requires atleast one attribute value to exist. key:%s, numfound:%d",key,numfound);
		gconf_set_error(err, GCONF_ERROR_BAD_KEY, 
			_("Key has wrong number of attribute values, we require atleast one attribute value to exist.. Key `%s'"), 
			dn);
		g_free(dn);
		return NULL;
	}
	ret = g_strv_copy( vals );
	ldap_value_free(vals);

	return ret;
}



/**
 * Given a valid "entry" from a search, this function looks up the attribute using
 * "key" and returns the value for that attribute. Note that the function will 
 * set an error (and return NULL) in case of any problems.
 * 
 * The function assumes that if the attribute is not available, or if there are
 * many values for the attribute then an error has occured.
 *
 * ld_source Instance data for backend
 * err Where to report errors to
 * entry Results from some other search
 * key Name of attribute to lookup
 * 
 * Caller must g_free() the return value.
 */
gchar* 
_ldap_gconf_get_value( LDAP* ld, GError** err, LDAPMessage* entry, gchar* key ) 
{
	char  **vals;
	gchar* ret;
	int numfound = 0;

	/* Determine the datatype of the key */
	vals = ldap_get_values( ld, entry, key );
	numfound = ldap_count_values(vals);
	if( 1 != numfound ) {
		gchar *dn = ldap_get_dn(ld, entry);
		gconf_log(	GCL_DEBUG, "_ldap_gconf_get_value NOT JUST ONE ATTRIB VALUE. key:%s, numfound:%d",key,numfound);
		gconf_set_error(err, GCONF_ERROR_BAD_KEY, 
			_("Key has wrong number of attribute values. Key `%s'"), 
			dn);
		g_free(dn);
		return NULL;
	}
	ret = g_strdup( vals[0] );
	ldap_value_free(vals);

	return ret;
}




/**
 * Given an LDAP entry from the directory, reify a GConfValue from that data.
 * In other words, given all the attributes for an entry, make a GConfValue.
 *
 * Note that you may wish to call _ldap_gconf_get_GConf_Value() if you have a
 * search reault because _ldap_gconf_get_GConf_Value() will check that you
 * have only one entry etc.
 *
 *
 * ld_source Instance data for backend
 * err Where to report errors to
 * entry The data for the new GConfValue
 *
 * if NULL is returned then an error is also set.
 */
GConfValue*
ldap_gconf_reify_gvalue( 
	const LDAPSource*const  ld_source, 
	GError** err, 
	LDAPMessage* entry
	)
{
	LDAP*   	ld 		= ld_source->ld;	
	GConfValue* ret 	= NULL;
	gchar*		dn		= NULL;
	struct read_data rdd;
	static deserialize_to_gconf_value_vtable vtab = {
		DESERIALIZE_TO_GCONF_VALUE_VTABLE_VERSION_1,
		read_type_f,
		r_val_f,
		r_pair_f,
		r_list_f,
		r_marshaled_f
	};
	gconf_log(	GCL_DEBUG, "ldap_gconf_reify_gvalue called");

	rdd.ld_source = ld_source;
	rdd.e = entry;
	
	dn = ldap_get_dn( ld, entry );
	
	ret = deserialize_to_gconf_value( 
		err,
		dn,
		ld_source->serialize_as,
		&rdd,
		&vtab);	
	free( dn );
	gconf_log(	GCL_DEBUG, "ldap_gconf_reify_gvalue done");
	
	return ret;
}



/**
 * Given a message response, find the GConf key that is serialized at this dn and return it as
 * a correctly typed GConfValue. If there is an error then error information is set in "err" and
 * NULL is returned. 
 *
 * Note that res is not free'd by this function, so that the caller may try other things after 
 * this call. It does mean that the caller still requires to ldap_msgfree(res) before it returns though.
 *
 * This function should be called when there is the possibiliy that there is more than
 * one entry (an error) rather than ldap_gconf_reify_gvalue(). If you are sure that
 * you have the one and only entry you can avoid the error handling of this function by
 * simply calling ldap_gconf_reify_gvalue()
 *
 * ld_source Instance data for backend
 * err Where to report errors to
 * res The raw results of a search
 * retval Where to put the goods if err is not set.
 *
 */
GConfValue*
_ldap_gconf_get_GConf_Value( 
	const LDAPSource*const  ld_source, 
	GError** err, 
	LDAPMessage* res,
	GConfValue** retval)
{
	int numfound = 0;
	LDAPMessage     *entry;
	LDAP*   	ld = ld_source->ld;	


	gconf_log(	GCL_DEBUG, "_ldap_gconf_get_GConf_Value called");

	
	/* did we get anything ? */
	if ((numfound = ldap_count_entries( ld_source->ld, res )) == -1) {
		gchar *dn = ldap_get_dn(ld, res);
		gconf_log(	GCL_DEBUG, "_ldap_gconf_get_GConf_Value didn't get anything");
		gconf_set_error(err, GCONF_ERROR_BAD_KEY, 
			_("Failed to find any data for key `%s'"), dn);
		g_free(dn);
		return NULL;
	}

	/* Was there too much data there, makes sense to report an error in this case right? */
	if( numfound != 1) {
		gchar *dn = ldap_get_dn(ld, res);
		gconf_log(	GCL_DEBUG, "_ldap_gconf_get_GConf_Value not one %d",numfound);
		gconf_set_error(err, GCONF_ERROR_BAD_KEY, 
			_("Too many/few entries in LDAP directory for given key `%s'"), dn);
		g_free(dn);
		return NULL;
	}

	/* There is only one entry */
	entry = ldap_first_entry( ld, res );
	*retval = ldap_gconf_reify_gvalue( ld_source, err, entry );

	gconf_log(	GCL_DEBUG, "_ldap_gconf_get_GConf_Value done");
	
	return *retval;	
}


/**
 * Create a single entry in LDAP, returning the LDAP return code/error.
 *
 * ld LDAP connection
 * dn The dn for the new entry
 *
 * Returns the LDAP error code for this operation.
 */
int
ldap_gconf_create_ldap_entry(LDAP* ld, const gchar* dn)
{
	GPtrArray* a;
	int ldap_err;

	if((a=ldap_gconf_ldapmodvec_new())) {
		ldap_gconf_ldapmodvec_addmod(a, 0, "objectClass", "externalObject", NULL);
		ldap_err = ldap_add_s(ld, dn, ldap_gconf_ldapmodvec_toArray(a));
		ldap_gconf_ldapmodvec_free(a);
		return ldap_err; 
	}
	return LDAP_NO_MEMORY; /* kinda hackish in some light? */
}


/**
 * Remove key_base from the given dn 
 *
 * This is a handy function every now and then because _key_to_ldap_dn() makes the 
 * fully qualified dn, which includes the key_base.
 *
 * Caller must free return string.
 */
gchar*
ldap_gconf_remove_base_dn( const LDAPSource *const ld_source, const gchar* dn ) 
{
	gchar* tmp;
	tmp = g_strndup( dn, strlen(dn) - strlen(ld_source->key_base));
	tmp = g_strchomp (tmp);
	if( tmp[strlen(tmp)-1] && tmp[strlen(tmp)-1] == ',') {
		tmp[strlen(tmp)-1] = ' ';
	}
	return tmp;
}





/**
 *
 * Creates the LDAP tree (if it doesn't exist already) right down to the level of dn. 
 *
 * Basic algorithmn if(check) then there may be no need to make the tree, if this is the
 * case then we try to add the full given dn, if there is a EXISTS error then we can
 * save alot of network IO by simply returning. Otherwise:
 *
 * Rip apart the given dn, starting from the smallest right most entry trying to add
 * each item, if there is an error other than a EXISTS type then we failed, err is set and 1 
 * is returned. Otherwise we keep moving back to the full given dn until all of the tree is
 * created. Then we are done.
 *
 * Items in key_base are not created, they are assumed to be static, eg. the root of the gnome
 * data in the LDAP tree. 
 *
 * ld_source Instance data
 * err Error place
 * dn The dn of the tree that might need to be created (may partialy/fully exist already)
 * check Should the function check to see if the tree is already there first.
 *
 * returns 1 if there is an error (and err is set) otherwise returns 0
 */
int
ldap_gconf_create_ldap_entry_tree(
	LDAPSource* ld_source, GError** err, const gchar* dn, gboolean check)
{
	LDAP* ld = ld_source->ld;
	int i=0;
	gchar** vec;
	gchar* tmp;
	int ldap_err;
	
	gconf_log(	GCL_DEBUG, "ldap_gconf_create_ldap_entry_tree top dn:%s ld_source->key_base:%s",dn,ld_source->key_base);

	tmp = ldap_gconf_remove_base_dn(ld_source, dn);
	vec =  g_strsplit(tmp,",",0);
	g_free(tmp);

	i = g_strv_length( vec );
	gconf_log(	GCL_DEBUG, "ldap_gconf_create_ldap_entry_tree count(i) %d",i);
	
	for( i--; i >= 0; --i) {
		gchar* rdn;
		gchar* base_plus_rdn;
		
		rdn = g_strjoinv( "," , &vec[i] );
		base_plus_rdn = g_strconcat( rdn, ",", ld_source->key_base, NULL );
		gconf_log(	GCL_DEBUG, "ldap_gconf_create_ldap_entry_tree create rdn:%s",rdn);
		gconf_log(	GCL_DEBUG, "ldap_gconf_create_ldap_entry_tree create base_plus_rdn:%s",base_plus_rdn);
		
		ldap_err = ldap_gconf_create_ldap_entry(ld, base_plus_rdn);
		g_free(rdn);
		
		if( ldap_err != LDAP_SUCCESS 
			&& ldap_err != LDAP_TYPE_OR_VALUE_EXISTS
			&& ldap_err != LDAP_ALREADY_EXISTS )
		{
			gconf_log(	GCL_DEBUG, "ldap_gconf_create_ldap_entry_tree bailing ldap_err:%s",ldap_err2string(ldap_err));
			_ldap_gconf_set_error(err, GCONF_ERROR_FAILED, ldap_err, 
				_("Can not make dn `%s'"), base_plus_rdn);
			g_free(base_plus_rdn);
			g_strfreev(vec);
			return 1;
		}
		g_free(base_plus_rdn);
	}
	
	g_strfreev(vec);
	return 0;
}

/**
 * Execute stored modifications on the LDAP server.
 *
 * If an error occurs then err is set and 1 returned, if 0 returned then all is ok
 *
 * ld_source Instance data
 * err Error place
 * dn The dn to set attr=value for
 * a The series of modifications to perform
 *
 */
int
_ldap_gconf_execute_mods( LDAPSource* ld_source, GError** err, const gchar* dn, GPtrArray* a )
{
	int ldap_err = 0;
	LDAP* ld = ld_source->ld;
	
	if((ldap_err=ldap_modify_s(ld, dn, ldap_gconf_ldapmodvec_toArray(a))) != LDAP_SUCCESS)
	{
		/* Create the whole tree for this entry */
		if(LDAP_NO_SUCH_OBJECT == ldap_err) {
			gconf_log(	GCL_DEBUG, "_ldap_gconf_execute_mods Must create tree for dn:%s",dn);
			if(ldap_gconf_create_ldap_entry_tree( ld_source, err, dn, 0)) {
				gconf_log(	GCL_DEBUG, "_ldap_gconf_execute_mods COULD NOT CREATE TREE dn:%s",dn);
				return 1;
			}
			if((ldap_err=ldap_modify_s(ld, dn, ldap_gconf_ldapmodvec_toArray(a))) == LDAP_SUCCESS) {
				gconf_log(	GCL_DEBUG, "_ldap_gconf_execute_mods all is ok after tree create");
				return 0;
			}
		}
		
		gconf_log(	GCL_DEBUG, "_ldap_gconf_execute_mods error happened %d, %s",ldap_err,ldap_err2string(ldap_err));
		_ldap_gconf_set_error( err, GCONF_ERROR_BAD_KEY, ldap_err, "",NULL);
		return 1;
	}
	return 0;
}


/**
 * Counts the number of LDAP name components in dn
 * 
 * eg. "dn=me, dn=there" == 2
 */
int
_ldap_count_dn_components( gchar* dn ) 
{
	int r = 0;
	int i = 0;
	if( !dn || !dn[0]) return 0;
		
	for( i=0; dn[i]; i++) {
		if( dn[i] == ',' ) {
			r++;
		}
	}
	return r+1;
}




/**
 * Use this function to wipe out all the data in "res". This depends on you having a legal
 * bunch of data in "res". eg. A whole subtree is fine, but if you try to delete a part of
 * a subtree, only the lowest part will work, if there is a dn in "res" that has a dn that
 * is longer than those in "res" this call will fail.
 *
 * eg. a call to delete b and c will fail due to d needing b to exist. It is up to the
 * caller to make sure a full subtree is in "res". (use ldap_search to do this easily)
 * a
 * \b
 *  \c
 *  \d
 *
 *
 * Algo: step through each entry returned and break it up by its dn, what we
 * want is a hash of lists. each list will have dns that have the same number
 * of components in them, so that we can just start with the list that has the
 * longest relative dn (rdn), and delete them all, then this will let us delete the list
 * with the second longest rdn because there are no longer entries handing of that rdn.
 * 
 * returns 1 on error, and err is set. returns 0 if ok.
 */
int
_ldap_gonf_delete_recursive( const LDAPSource* const ld_source, GError** err, LDAPMessage* res )
{
	LDAPMessage* e;
	GHashTable* hash = g_hash_table_new(0,0);
	int ret = 0;
	int max = 0;
	int i = 0;
	int ldap_err = 0;
	gchar* dn;
	
	gconf_log(	GCL_DEBUG, "_ldap_gonf_delete_recursive called");

	/* sort into array of lists */
	for ( e = ldap_first_entry( ld_source->ld, res ); e != NULL;
		e = ldap_next_entry( ld_source->ld, e ) ) 
	{
		guint count;
		GBlob* blob;
				
		dn = ldap_get_dn( ld_source->ld, e );
		gconf_log(	GCL_DEBUG, "_ldap_gonf_delete_recursive  got dn:%s",dn);
		
		/* count how long this dn is (in terms of components in its name, and
		   place it in the proper list 
		*/
		gconf_log(	GCL_DEBUG, "_ldap_gonf_delete_recursive  1");
		count = _ldap_count_dn_components(dn);
		gconf_log(	GCL_DEBUG, "_ldap_gonf_delete_recursive count:%d",count);
		
		
		gconf_log(	GCL_DEBUG, "_ldap_gonf_delete_recursive  ---A");
		blob = g_hash_table_lookup( hash, GINT_TO_POINTER(count) );
		gconf_log(	GCL_DEBUG, "_ldap_gonf_delete_recursive  ---B");
		
		if( !blob ) {
			blob = g_blob_new();
			g_blob_push(blob, g_strdup(dn));
			gconf_log(	GCL_DEBUG, "_ldap_gonf_delete_recursive  adding new list");
			g_hash_table_insert (hash, GINT_TO_POINTER(count), blob);
			gconf_log(	GCL_DEBUG, "_ldap_gonf_delete_recursive  added new list");
			if( max < count ) {
				max = count;
			}
		}
		else 
		{
			gconf_log(	GCL_DEBUG, "_ldap_gonf_delete_recursive  2");
			g_blob_push( blob, g_strdup(dn));
			gconf_log(	GCL_DEBUG, "_ldap_gonf_delete_recursive  3");
		}		
		free( dn );
	}			
	gconf_log(	GCL_DEBUG, "_ldap_gonf_delete_recursive  ---about to start delete them all");
	
	/*
	* Now delete the entries in the lists, largest one first 
	*/
	for( i=max; i >=0 ; i-- ) {
		GBlob* blob;
		
		blob = g_hash_table_lookup( hash, GINT_TO_POINTER(i) );
		
		while((dn = g_blob_pop( blob ))) 
		{
			/* We have this here so that we can keep going in case of error,
				but only free the data, quite handy really
			*/
			if( !ret ) 
			{
				gconf_log(	GCL_DEBUG, "element is:%s\n",dn);
	
				if(LDAP_SUCCESS != (ldap_err = ldap_delete_s( ld_source->ld, dn ))) {
					gconf_log(	GCL_DEBUG, "_ldap_gconf_delete failed on %s reason %s",dn,
						ldap_err2string(ldap_err));
					_ldap_gconf_set_error( err, GCONF_ERROR_FAILED, ldap_err, 
						"Problem with deletion for dn: %s",dn);

					ret = 1; 
				}
			}			
			g_free(dn);
		}		
		
		gconf_log(	GCL_DEBUG, "_ldap_gonf_delete_recursive  ---glistfree1");
		g_free (blob);
		gconf_log(	GCL_DEBUG, "_ldap_gonf_delete_recursive  ---glistfree2");
	}

	gconf_log(	GCL_DEBUG, "_ldap_gonf_delete_recursive  ---nearing the end!");
	g_hash_table_destroy(hash);
	
	gconf_log(	GCL_DEBUG, "_ldap_gonf_delete_recursive _ret %d",ret);
	return ret;
}

#include <time.h> /* only for _convert_from_ldap_time */
/**
 * Convert an uTCTimeSyntax time to a GTime format.
 *
 * Note that this is a hack, I have no doco on the uTCTimeSyntax. I don't wish to pay for it
 * so I am guessing how to do it FIXME (uTCTimeSyntax)
 *
 *
 * 012345678901234567 <-- index data
 *     012345678901234567
 *       012345678901234567
 *         012345678901234567
 *           012345678901234567
 *             012345678901234567
 * 20001114105036Z <-- a LDAP time
 * 2000.11.14...10.50.36.Z"; <-- a LDAP time broken down
 * 
 *
 * Returns a value time or -1.
 */
GTime
_convert_from_ldap_time( gchar* ldapt )
{
	gchar* tstr;
	struct tm stm;
	time_t tt=0;
		

	tstr = g_strdup(ldapt);
	tstr[4] = '\0';
	stm.tm_year = atoi(tstr)-1900; g_free(tstr);
	
	tstr = g_strdup(&ldapt[4]);
	tstr[2] = '\0';
	stm.tm_mon = atoi(tstr)-1; g_free(tstr);
	
	tstr = g_strdup(&ldapt[6]);
	tstr[2] = '\0';
	stm.tm_mday = atoi(tstr); g_free(tstr);

	tstr = g_strdup(&ldapt[8]);
	tstr[2] = '\0';
	stm.tm_hour = atoi(tstr); g_free(tstr);
	
	tstr = g_strdup(&ldapt[10]);
	tstr[2] = '\0';
	stm.tm_min = atoi(tstr); g_free(tstr);

	tstr = g_strdup(&ldapt[12]);
	tstr[2] = '\0';
	stm.tm_sec = atoi(tstr); g_free(tstr);


	////////// some test data for mktime ///////////
/*
	stm.tm_year = 100; // -1900
	stm.tm_mon = 10; // -1
	stm.tm_mday = 14;
	stm.tm_hour = 11;
	stm.tm_min = 12;
	stm.tm_sec = 1;
*/

	stm.tm_wday = 0; /* ignored */
	stm.tm_yday = 0; /* ignored */
	stm.tm_isdst = 0; /* hmm ? */
	
	tt = mktime( &stm );

	return tt;
}


/**
 * Given the standard GConf key of /something/else/file
 * create the GConf key for the meta data. Note the returned key will still
 * be in filesystem like format, not LDAP format.
 *
 * Returns NULL on error, or a string that the caller must free. 
 */
gchar*
_ldap_gconf_make_metakey( const gchar* key )
{
	return g_strconcat( key, LDAP_GCONF_META_SCHEMA_POSTFIX, NULL);
}



/** 
 * Call here when you have a string alloc error etc. ie. out of RAM.
 */
void
_ldap_err_no_ram( GError** err ) 
{
	gconf_set_error( err, GCONF_ERROR_FAILED, "Out of RAM");
}







void 
_ldap_gconf_type_and_val_f( 
	GError** err, 
	const gchar* key, 
	const gchar* type, 
	gchar* value,
	gpointer user_data)
{
	LDAPSource* ld_source = (LDAPSource*)user_data;
	LDAP* ld = ld_source->ld;
	GPtrArray* a;

	a=ldap_gconf_ldapmodvec_new();
	ldap_gconf_ldapmodvec_addmod(a, LDAP_MOD_REPLACE, LDAP_GCONF_TYPE_KEY, type, NULL);
	ldap_gconf_ldapmodvec_addmod(a, LDAP_MOD_REPLACE, LDAP_GCONF_VALUE_KEY, value, NULL);
	ldap_gconf_ldapmodvec_addmod(a, LDAP_MOD_REPLACE, LDAP_GCONF_OBJECTCLASS_KEY, LDAP_GCONF_OBJECTCLASS, NULL);
	_ldap_gconf_execute_mods( ld_source, err, key, a );	
	ldap_gconf_ldapmodvec_free(a);
	return;
}

void 
_ldap_gconf_pair_f( 
	GError** err, 
	const gchar* key, 
	const gchar* type, 
	const gchar* car_type, 
	const gchar* cdr_type, 
	gchar* car_value, 
	gchar* cdr_value,
	gpointer user_data)
{
	LDAPSource* ld_source = (LDAPSource*)user_data;
	LDAP* ld = ld_source->ld;
	GPtrArray* a;

	a=ldap_gconf_ldapmodvec_new();
	ldap_gconf_ldapmodvec_addmod(a, LDAP_MOD_REPLACE, LDAP_GCONF_TYPE_KEY, type, NULL);
	ldap_gconf_ldapmodvec_addmod(a, LDAP_MOD_REPLACE, LDAP_GCONF_CAR_TYPE_KEY, car_type, NULL);
	ldap_gconf_ldapmodvec_addmod(a, LDAP_MOD_REPLACE, LDAP_GCONF_CDR_TYPE_KEY, cdr_type, NULL);
	ldap_gconf_ldapmodvec_addmod(a, LDAP_MOD_REPLACE, LDAP_GCONF_CAR_VALUE_KEY, car_value, NULL);
	ldap_gconf_ldapmodvec_addmod(a, LDAP_MOD_REPLACE, LDAP_GCONF_CDR_VALUE_KEY, cdr_value, NULL);
	ldap_gconf_ldapmodvec_addmod(a, LDAP_MOD_REPLACE, LDAP_GCONF_OBJECTCLASS_KEY, LDAP_GCONF_OBJECTCLASS, NULL);
	_ldap_gconf_execute_mods( ld_source, err, key, a );	
	ldap_gconf_ldapmodvec_free(a);
	return;
}

void 
_ldap_gconf_list_f( 
	GError** err, 
	const gchar* key, 
	const gchar* type, 
	char** values,
	gpointer user_data)
{
	LDAPSource* ld_source = (LDAPSource*)user_data;
	LDAP* ld = ld_source->ld;
	GPtrArray* a;

	a=ldap_gconf_ldapmodvec_new();
	ldap_gconf_ldapmodvec_addmod (a, LDAP_MOD_REPLACE, LDAP_GCONF_TYPE_KEY, type, NULL);
	ldap_gconf_ldapmodvec_addmodv(a, LDAP_MOD_REPLACE, LDAP_GCONF_LIST_VALUE_KEY, values);
	ldap_gconf_ldapmodvec_addmod (a, LDAP_MOD_REPLACE, LDAP_GCONF_OBJECTCLASS_KEY, LDAP_GCONF_OBJECTCLASS, NULL);
	_ldap_gconf_execute_mods( ld_source, err, key, a );	
	ldap_gconf_ldapmodvec_free(a);
	return;
}

void 
_ldap_gconf_marshaled_f( 
	GError** err, 
	const gchar* key, 
	char* value,
	gpointer user_data)
{
	LDAPSource* ld_source = (LDAPSource*)user_data;
	LDAP* ld = ld_source->ld;
	GPtrArray* a;

	a=ldap_gconf_ldapmodvec_new();
	ldap_gconf_ldapmodvec_addmod(a, LDAP_MOD_REPLACE, LDAP_GCONF_TYPE_KEY, LDAP_GCONF_TYPE_MARSHALED, NULL);
	ldap_gconf_ldapmodvec_addmod(a, LDAP_MOD_REPLACE, LDAP_GCONF_MARSHALED_VALUE_KEY, value, NULL);
	ldap_gconf_ldapmodvec_addmod(a, LDAP_MOD_REPLACE, LDAP_GCONF_OBJECTCLASS_KEY, LDAP_GCONF_OBJECTCLASS, NULL);
	_ldap_gconf_execute_mods( ld_source, err, key, a );	
	ldap_gconf_ldapmodvec_free(a);
	return;
}



gchar* 
read_type_f( 
	GError** err, const gchar* key, gpointer user_data )
{
	struct read_data* rdd = (struct read_data*)user_data;
	LDAPSource* ld_source = rdd->ld_source;
	LDAP* ld = ld_source->ld;
	LDAPMessage** e = &rdd->e;
	
	return _ldap_gconf_get_value( ld, err, *e, LDAP_GCONF_TYPE_KEY );
}

GConfValue* 
r_val_f( GError** err, const gchar* key, gchar* type, gpointer user_data )
{
	struct read_data* rdd = (struct read_data*)user_data;
	LDAPSource* ld_source = rdd->ld_source;
	LDAP* ld = ld_source->ld;
	LDAPMessage** e = &rdd->e;
	gchar* value = NULL;
	GConfValue* ret = NULL;
	
	value = _ldap_gconf_get_value( ld, err, *e, LDAP_GCONF_VALUE_KEY );
	if(!value) {
		return NULL;
	}

	ret = gconf_value_new_from_string( 
			gconf_value_type_from_string(type),
			value,
			err);
	
	g_free( value );
	return ret;
}

GConfValue* 
r_pair_f( 
	GError** err, const gchar* key, gchar* type, gpointer user_data )
{
	
	struct read_data* rdd = (struct read_data*)user_data;
	LDAPSource* ld_source = rdd->ld_source;
	LDAP* ld 			= ld_source->ld;
	LDAPMessage** e 	= &rdd->e;
	gchar* car_value	= NULL;
	gchar* cdr_value	= NULL;
	gchar* car_type		= NULL;
	gchar* cdr_type 	= NULL;
	GConfValue* ret 	= NULL;
	gint error			= 0;
	
	if((car_type  = _ldap_gconf_get_value( ld, err, *e, LDAP_GCONF_CAR_TYPE_KEY ))) {
		if((cdr_type  = _ldap_gconf_get_value( ld, err, *e, LDAP_GCONF_CDR_TYPE_KEY ))) {
			if((car_value = _ldap_gconf_get_value( ld, err, *e, LDAP_GCONF_CAR_VALUE_KEY ))) {
				if((cdr_value = _ldap_gconf_get_value( ld, err, *e, LDAP_GCONF_CDR_VALUE_KEY ))) {
					GConfValue *t = NULL;

	    			ret = gconf_value_new (GCONF_VALUE_PAIR);
					if( ret ) {
						
						t = gconf_value_new_from_string( 
							gconf_value_type_from_string(car_type),
							car_value,
							err);					
						if(t) {
							gconf_value_set_car_nocopy (ret, t);
							t = gconf_value_new_from_string( 
								gconf_value_type_from_string(cdr_type),
								cdr_value,
								err);					
							if(t) {
								gconf_value_set_cdr_nocopy (ret, t);
							}
							else {
								error = 1;
							}
						}
						else {
							error = 1;
						}
					}
				}
			}
		}
	}
	
	if( car_value ) 	g_free(car_value);
	if( cdr_value ) 	g_free(cdr_value);
	if( car_type ) 		g_free(car_type);
	if( cdr_type ) 		g_free(cdr_type);
	if( error )			{
		if(ret) {
			gconf_value_free ( ret );
			ret = NULL;
		}
	}
	
	return ret;
}

/** Added by Ben Martin */
GConfValue*
gconf_value_set_list_strv_nocopy(gchar** strv)
{
	GSList *list 		= NULL;
	int i 				= 0;
  	GConfValue* value 	= NULL;
	
  	if( strv ) {
		for( i=0; strv[i]; i++ )
			list = g_slist_append(list, strv[i]);
	}
	
	value = gconf_value_new(GCONF_VALUE_LIST);
	gconf_value_set_list_type(value, GCONF_VALUE_STRING);
	gconf_value_set_list_nocopy(value,list);
	
	return value;
}

	
GConfValue* 
r_list_f( 
	GError** err, const gchar* key, gchar* type, gpointer user_data )
{
	struct read_data* rdd = (struct read_data*)user_data;
	LDAPSource* ld_source = rdd->ld_source;
	LDAP* ld 			= ld_source->ld;
	LDAPMessage** e 	= &rdd->e;
	gchar** vals		= NULL;
	GConfValue* ret 	= NULL;
	
	vals = _ldap_gconf_get_values( ld, err, *e, LDAP_GCONF_LIST_VALUE_KEY );
	if( vals ) 
	{
		ret = gconf_value_set_list_strv_nocopy(vals);
	}
	
	return ret;
}

gchar* 
r_marshaled_f( 
	GError** err, const gchar* key, gpointer user_data )
{
	struct read_data* rdd = (struct read_data*)user_data;
	LDAPSource* ld_source = rdd->ld_source;
	LDAP* ld = ld_source->ld;
	LDAPMessage** e = &rdd->e;
	
	return _ldap_gconf_get_value( ld, err, *e, LDAP_GCONF_MARSHALED_VALUE_KEY );
}

