
/*
 * ldap gconf backend
 *
 * Copyright (C) 2000 Ben Martin (aka monkeyiq)
 *
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

// ldap
#include <ldap.h> 

// gconf
#include <gconf/gconf-backend.h>
#include <gconf/gconf-internals.h>
#include <gconf/gconf.h>

// glib
#include <glib.h>

// local
#include "ldap-backend.h"

/**
 * This function tags on LDAP specific error information to the GError that is passed back
 * to the user.
 *
 * ldap_err is the LDAP error code that occured.
 *
 * all other params are the same as gconf_set_error()
 */
void
_ldap_gconf_set_error(GError** err, GConfError en, int ldap_err, const gchar* fmt, ...);

/**
 * Debug code that dumps out a char** to the gconf log fle.
 *
 * vec The vector of strings to log.
 */
void
g_strv_show( gchar** vec );

/**
 * reverse a gchar**
 *
 * Note that the input vector of strings is moved in place. There is the 
 * expectation that "vec" is NULL terminated.
 *
 */
gchar** g_strv_reverse( gchar ** vec );

/**
 * Construct a LDAP DN for a given GConf path. 
 *
 * If an error occurs "err" is set and NULL returned.
 *
 * ld_source Instance data for backend
 * err Where to report errors to
 * gk The gconf key to translate into a LDAP dn.
 *
 * returned is a string that is the LDAP dn, caller must g_free() return value.
 */
gchar* 
_key_to_ldap_dn( const LDAPSource* const  ld_source, GError** err, const gchar* gk);


/**
 * Perform a ldap_search() on the database, if there is an error, use "err" to report that
 * to the user, and return 1.
 *
 * The return value from this function is 1 if there is an error, in which case the caller 
 * should return NULL to the gconf engine.
 *
 * err The gconf err to put error information into
 * scope is the LDAP scope to search for, if 0 then it is BASE (ie exact match)
 * ldap_key The dn to lookup
 * res Where to place the attributes for this key.
 */
gint
_ldap_gconf_lookup_key( 
	const LDAPSource* const ld_source, 
	GError** err, 
	int scope,
	gchar *filter,
	gchar** attrs,
	gchar* ldap_key, 
	LDAPMessage** res );


/**
 * Given a valid "entry" from a search, this function looks up the attribute using
 * "key" and returns the value for that attribute. Note that the function will 
 * set an error (and return NULL) in case of any problems.
 * 
 * The function assumes that if the attribute is not available, or if there are
 * many values for the attribute then an error has occured.
 *
 * ld_source Instance data for backend
 * err Where to report errors to
 * entry Results from some other search
 * key Name of attribute to lookup
 * 
 * Caller must g_free() the return value.
 */
gchar* 
_ldap_gconf_get_value( LDAP* ld, GError** err, LDAPMessage* entry, gchar* key );

/**
 * Given a valid "entry" from a search, this function looks up the attribute using
 * "key" and returns the values for that attribute. Note that the function will 
 * set an error (and return NULL) in case of any problems.
 * 
 * The function assumes that if the attribute is not available then an error has occured.
 *
 * ld_source Instance data for backend
 * err Where to report errors to
 * entry Results from some other search
 * key Name of attribute to lookup
 * 
 * Caller must g_strfreev() the return value.
 */
gchar** 
_ldap_gconf_get_values( LDAP* ld, GError** err, LDAPMessage* entry, gchar* key );

/**
 * Given an LDAP entry from the directory, reify a GConfValue from that data.
 * In other words, given all the attributes for an entry, make a GConfValue.
 *
 * Note that you may wish to call _ldap_gconf_get_GConf_Value() if you have a
 * search reault because _ldap_gconf_get_GConf_Value() will check that you
 * have only one entry etc.
 *
 *
 * ld_source Instance data for backend
 * err Where to report errors to
 * entry The data for the new GConfValue
 *
 * if NULL is returned then an error is also set.
 */
GConfValue*
ldap_gconf_reify_gvalue( 
	const LDAPSource*const  ld_source, 
	GError** err, 
	LDAPMessage* entry
	);


/**
 * Given a message response, find the GConf key that is serialized at this dn and return it as
 * a correctly typed GConfValue. If there is an error then error information is set in "err" and
 * NULL is returned. 
 *
 * Note that res is not free'd by this function, so that the caller may try other things after 
 * this call. It does mean that the caller still requires to ldap_msgfree(res) before it returns though.
 *
 * This function should be called when there is the possibiliy that there is more than
 * one entry (an error) rather than ldap_gconf_reify_gvalue(). If you are sure that
 * you have the one and only entry you can avoid the error handling of this function by
 * simply calling ldap_gconf_reify_gvalue()
 *
 * ld_source Instance data for backend
 * err Where to report errors to
 * res The raw results of a search
 * retval Where to put the goods if err is not set.
 *
 */
GConfValue*
_ldap_gconf_get_GConf_Value( 
	const LDAPSource*const  ld_source, 
	GError** err, 
	LDAPMessage* res,
	GConfValue** retval);

/**
 * Create a single entry in LDAP, returning the LDAP return code/error.
 *
 * ld LDAP connection
 * dn The dn for the new entry
 *
 * Returns the LDAP error code for this operation.
 */
int
ldap_gconf_create_ldap_entry(LDAP* ld, const gchar* dn);


/**
 * Remove key_base from the given dn 
 *
 * This is a handy function every now and then because _key_to_ldap_dn() makes the 
 * fully qualified dn, which includes the key_base.
 *
 * Caller must free return string.
 */
gchar*
ldap_gconf_remove_base_dn( const LDAPSource *const ld_source, const gchar* dn );

/**
 *
 * Creates the LDAP tree (if it doesn't exist already) right down to the level of dn. 
 *
 * Basic algorithmn if(check) then there may be no need to make the tree, if this is the
 * case then we try to add the full given dn, if there is a EXISTS error then we can
 * save alot of network IO by simply returning. Otherwise:
 *
 * Rip apart the given dn, starting from the smallest right most entry trying to add
 * each item, if there is an error other than a EXISTS type then we failed, err is set and 1 
 * is returned. Otherwise we keep moving back to the full given dn until all of the tree is
 * created. Then we are done.
 *
 * Items in key_base are not created, they are assumed to be static, eg. the root of the gnome
 * data in the LDAP tree. 
 *
 * ld_source Instance data
 * err Error place
 * dn The dn of the tree that might need to be created (may partialy/fully exist already)
 * check Should the function check to see if the tree is already there first.
 *
 * returns 1 if there is an error (and err is set) otherwise returns 0
 */
int
ldap_gconf_create_ldap_entry_tree(
	LDAPSource* ld_source, GError** err, const gchar* dn, gboolean check);




/**
 * Counts the number of LDAP name components in dn
 * 
 * eg. "dn=me, dn=there" == 2
 */
int
_ldap_count_dn_components( gchar* dn );


/**
 * Use this function to wipe out all the data in "res". This depends on you having a legal
 * bunch of data in "res". eg. A whole subtree is fine, but if you try to delete a part of
 * a subtree, only the lowest part will work, if there is a dn in "res" that has a dn that
 * is longer than those in "res" this call will fail.
 *
 * eg. a call to delete b and c will fail due to d needing b to exist. It is up to the
 * caller to make sure a full subtree is in "res". (use ldap_search to do this easily)
 * a
 * \b
 *  \c
 *  \d
 *
 *
 * Algo: step through each entry returned and break it up by its dn, what we
 * want is a hash of lists. each list will have dns that have the same number
 * of components in them, so that we can just start with the list that has the
 * longest relative dn (rdn), and delete them all, then this will let us delete the list
 * with the second longest rdn because there are no longer entries handing of that rdn.
 * 
 * returns 1 on error, and err is set. returns 0 if ok.
 */
int
_ldap_gonf_delete_recursive( const LDAPSource* const ld_source, GError** err, LDAPMessage* res );



/**
 * Convert an uTCTimeSyntax time to a GTime format.
 *
 * Note that this is a hack, I have no doco on the uTCTimeSyntax. I don't wish to pay for it
 * so I am guessing how to do it FIXME (uTCTimeSyntax)
 *
 *
 * 012345678901234567 <-- index data
 *     012345678901234567
 *       012345678901234567
 *         012345678901234567
 *           012345678901234567
 *             012345678901234567
 * 20001114105036Z <-- a LDAP time
 * 2000.11.14...10.50.36.Z"; <-- a LDAP time broken down
 * 
 *
 * Returns a value time or -1.
 */
GTime
_convert_from_ldap_time( gchar* ldapt );



/**
 * Given the standard GConf key of /something/else/file
 * create the GConf key for the meta data. Note the returned key will still
 * be in filesystem like format, not LDAP format.
 *
 * Returns NULL on error, or a string that the caller must free. 
 */
gchar*
_ldap_gconf_make_metakey( const gchar* key );


/** 
 * Call here when you have a string alloc error etc. ie. out of RAM.
 */
void
_ldap_err_no_ram( GError** err );

/**
 * get the length of a string vector
 *
 * vec The vector of strings
 */
int
g_strv_length( gchar** vec );


/**
 * Produce and return a verbatim copy of strv.
 *
 * Caller must g_strfreev() return value.
 */
gchar**
g_strv_copy( gchar** strv );


/**
 * Execute stored modifications on the LDAP server.
 *
 * If an error occurs then err is set and 1 returned, if 0 returned then all is ok
 *
 * ld_source Instance data
 * err Error place
 * dn The dn to set attr=value for
 * a The series of modifications to perform
 *
 */
int
_ldap_gconf_execute_mods( LDAPSource* ld_source, GError** err, const gchar* dn, GPtrArray* a );


/* serialize_from_gconf_value() callbacks */
void 
_ldap_gconf_type_and_val_f( 
	GError** err, 
	const gchar* key, 
	const gchar* type, 
	gchar* value,
	gpointer user_data);
	
void 
_ldap_gconf_pair_f( 
	GError** err, 
	const gchar* key, 
	const gchar* type, 
	const gchar* car_type, 
	const gchar* cdr_type, 
	gchar* car_value, 
	gchar* cdr_value,
	gpointer user_data);
	
void 
_ldap_gconf_list_f( 
	GError** err, 
	const gchar* key, 
	const gchar* type, 
	char** values,
	gpointer user_data);
	
void 
_ldap_gconf_marshaled_f( 
	GError** err, 
	const gchar* key, 
	char* value,
	gpointer user_data);

/* Data for all reading demux functions */
struct read_data {
	LDAPSource* ld_source;
	LDAPMessage* e;
};

gchar* 
read_type_f(GError** err, const gchar* key, gpointer user_data );

GConfValue* 
r_val_f( GError** err, const gchar* key, gchar* type, gpointer user_data );

GConfValue* 
r_pair_f(GError** err, const gchar* key, gchar* type, gpointer user_data );
	
GConfValue* 
r_list_f(	GError** err, const gchar* key, gchar* type, gpointer user_data );

gchar* 
r_marshaled_f(GError** err, const gchar* key, gpointer user_data );


