
/*
 * ldap backend shared code
 *
 * Copyright (C) 2000 Ben Martin (aka monkeyiq)
 *
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "serialize.h"

/**
 * Make a list into a string vector. Return value is a NULL terminated vector of strings.
 *
 * Note that gconf_value_to_string() could use this with g_strjoinv()
 *
 * Returned strvec should be freed by caller with g_strfreev(). Return value will never be NULL, it may
 * have 0 elements though, ie. *ret == NULL;
 */
char**
__gconf_list_to_strv( GConfValue* v ) {
	char** ret = NULL;
	GSList* list;

	gconf_log(	GCL_DEBUG, "__gconf_list_to_strv. called ");

	list = gconf_value_get_list(v);
	if (list == NULL) {
		ret = g_new0(char*,2);
		ret[0] = NULL;
		return ret;
	}
	else
	{
		GPtrArray* a = NULL;
		
		a = g_ptr_array_new ();
		g_assert(list != NULL);

		while (list != NULL) 
		{		
			gchar* tmp;
			
			gconf_log(	GCL_DEBUG, "__gconf_list_to_strv. 1 ");
			tmp = gconf_value_to_string((GConfValue*)list->data);
			gconf_log(	GCL_DEBUG, "__gconf_list_to_strv. 2 ");
			g_assert(tmp != NULL);
			
			g_ptr_array_add (a, tmp);
			list = g_slist_next(list);
		}
		
		g_ptr_array_add (a, NULL);
		ret = (gchar**) a->pdata;
		g_ptr_array_free( a, 0 );
	}
	return ret;
}

/**
 * demux the save operation, if the user wants some values to be saved as value & type
 * then demux that to the correct save function, otherwise, just marshal the data to
 * a single string and call a generic function to save that.
 *
 * Note that the default operation is the save the marshaled single string unless
 * the user has requested explicitly that values should be saved in a more
 * readable format.
 *
 * err Error place
 * key key to save
 * val Value to save
 * sat Special feild that contains user prefs for what way to save each value type.
 * 
 */
void
serialize_from_gconf_value( 
	GError** err, 
	const gchar* key, 
	GConfValue* value, 
	serialize_as_t sat,
	gpointer user_data,
	serialize_from_gconf_value_vtable* vtab)
{
	gboolean saved = 0;
	gchar *str = NULL;

#define ___call_type_and_val_f( satbit ) if(vtab->type_and_val_f && (sat & satbit)) { \
			gchar* str_val = gconf_value_to_string(value); \
			saved=1; \
			vtab->type_and_val_f( err, key, gconf_value_type_to_string(value->type), str_val, user_data); \
			g_free(str_val); \
		}
	
	/*
	 * Check to see if the user wants some more readable format for data.
	 */
	if ((sat & SERIALIZE_ANY_NO_MARSHAL)) {
	
		switch (value->type)
		{
		case GCONF_VALUE_INT: 		___call_type_and_val_f( SERIALIZE_INT_NO_MARSHAL ) break;
		case GCONF_VALUE_FLOAT:		___call_type_and_val_f( SERIALIZE_FLOAT_NO_MARSHAL ) break;
		case GCONF_VALUE_STRING:	___call_type_and_val_f( SERIALIZE_STRING_NO_MARSHAL ) break;
		case GCONF_VALUE_BOOL:		___call_type_and_val_f( SERIALIZE_BOOL_NO_MARSHAL ) break;
    	case GCONF_VALUE_PAIR:
			{
				if( vtab->pair_f && (sat & SERIALIZE_PAIR_NO_MARSHAL )) {
					gchar* car_s = gconf_value_to_string(gconf_value_get_car(value));
					gchar* cdr_s = gconf_value_to_string(gconf_value_get_cdr(value));

					saved = 1;
					vtab->pair_f( err, key, gconf_value_type_to_string( value->type ),
						gconf_value_type_to_string(gconf_value_get_car(value)->type),
						gconf_value_type_to_string(gconf_value_get_cdr(value)->type),
						car_s,
						cdr_s,
						user_data);
					g_free(car_s);
					g_free(cdr_s);
				}
			}
			break;
		case GCONF_VALUE_LIST:
			{
				if( vtab->list_f && (sat & SERIALIZE_LIST_NO_MARSHAL )) {
					gchar** strv = __gconf_list_to_strv( value );

					saved = 1;
					vtab->list_f( err, key, gconf_value_type_to_string( value->type ), strv, user_data);
					g_strfreev(strv);
				}
			}
			break;
		case GCONF_VALUE_INVALID:
		case GCONF_VALUE_SCHEMA:
			break;
		default:
			g_assert_not_reached();
			break;			
		}
	}

	/* 
	 * If the user has not explicitly requested that the value be saved in a human readable format
	 * then we marshal the data and save it as one single string.
	 */
	if( !saved ) {
		
		saved = 1;
		str = ___external_serialize_value_to_string( value );
		vtab->marshaled_f( err, key, str, user_data );
		g_free(str);
	}
	
	return;
}


/**
 * demux the load operation, if the user wants some values to be saved/loaded as value & type
 * then demux that to the correct load function, otherwise, just marshal the data to
 * a single string and call a generic function to load that.
 *
 * Note that the default operation is the load the marshaled single string unless
 * the user has requested explicitly that values should be saved in a more
 * readable format.
 *
 * Note that there is read_type_f here and there was no save_type_f in  serialize_from_gconf_value()
 * This exists both out of need and due to the fact that a load at this point is usually out of
 * RAM. If there were a save_type_f then there may need to be two lots of network IO in 
 * serialize_from_gconf_value().
 *
 * Note that the return value from read_type_f() and marshaled_f() are freed by this function.
 * 
 *
 * err Error place
 * key key to save
 * sat Special feild that contains user prefs for what way to save each value type.
 * 
 */
GConfValue*
deserialize_to_gconf_value( 
	GError** err, 
	const gchar* key, 
	serialize_as_t sat,
	gpointer user_data,
	deserialize_to_gconf_value_vtable* vtab
	)
{
	GConfValue* ret = NULL;
	gboolean loaded = 0;

#define ___call_f( satbit, func_name ) if(vtab->func_name && (sat & satbit)) { \
			loaded=1; \
			ret = vtab->func_name( err, key, type_str, user_data); \
		}
	
	/*
	 * Check to see if the user wants some more readable format for data.
	 */
	if ((sat & SERIALIZE_ANY_NO_MARSHAL) && vtab->read_type_f) {
	
		gchar* type_str = vtab->read_type_f( err, key, user_data );
		GConfValueType typev =  gconf_value_type_from_string( type_str );
		
		switch (typev)
		{
		case GCONF_VALUE_INT: 		___call_f( SERIALIZE_INT_NO_MARSHAL, val_f ) break;
		case GCONF_VALUE_FLOAT:		___call_f( SERIALIZE_FLOAT_NO_MARSHAL, val_f ) break;
		case GCONF_VALUE_STRING:	___call_f( SERIALIZE_STRING_NO_MARSHAL, val_f ) break;
		case GCONF_VALUE_BOOL:		___call_f( SERIALIZE_BOOL_NO_MARSHAL,val_f ) break;
    	case GCONF_VALUE_PAIR:		___call_f( SERIALIZE_PAIR_NO_MARSHAL,pair_f ) break;
		case GCONF_VALUE_LIST:		___call_f( SERIALIZE_LIST_NO_MARSHAL, list_f ) break;
		case GCONF_VALUE_INVALID:
		case GCONF_VALUE_SCHEMA:
			break;
		default:
			g_assert_not_reached();
			break;			
		}
	}

	/* 
	 * If the user has not explicitly requested that the value be saved in a human readable format
	 * then we marshal the data and save it as one single string.
	 */
	if( !loaded ) {
		gchar* str = NULL;
		
		loaded = 1;
		str = vtab->marshaled_f( err, key, user_data );
		ret = ___external_serialize_value_from_string( str );
		g_free(str);
	}
	
	return ret;
}
	

/**
 * Takes a GConfValue and creates a single string representation of it
 *
 * See also ___external_serialize_value_from_string()
 *
 * Caller must free return value.
 *
 */
char *
___external_serialize_value_to_string (GConfValue * val) {
	gchar* ret 	= NULL;
	gchar* tmp 	= NULL;
	size_t zz 	= 0;
	
	gconf_log(	GCL_DEBUG, "___external_serialize_value_to_string. called ");
	tmp = bdb_serialize_value (val, &zz);
	gconf_log(	GCL_DEBUG, "___external_serialize_value_to_string.");
	gconf_log(	GCL_DEBUG, "___external_serialize_value_to_string. string: %s",tmp);	
	ret = g_strdup(tmp);
	return ret;
}

/**
 * Creates a GConfValue from a single string representation of it what was
 * obtained using ___external_serialize_value_to_string()
 *
 * See also ___external_serialize_value_to_string()
 *
 * Caller must free return value.
 *
 */
GConfValue *
___external_serialize_value_from_string (const char *srz)
{
	return bdb_restore_value (srz);
}

