
/*
 * ldap backend shared code
 *
 * Copyright (C) 2000 Ben Martin (aka monkeyiq)
 *
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */
 
#ifndef INCLUDED_SERIALIZE_H
#define INCLUDED_SERIALIZE_H

#include <glib.h>
#include <gconf/gconf-backend.h>
#include <gconf/gconf-internals.h>
#include <gconf/gconf.h>

/* These define if we are going to store a string as a string value plus a type value,
   or if we are going to monster the value & type into one string and just save that
*/   
#define SERIALIZE_STRING_NO_MARSHAL	(1<<1)
#define SERIALIZE_INT_NO_MARSHAL	(1<<2)
#define SERIALIZE_BOOL_NO_MARSHAL	(1<<3)
#define SERIALIZE_FLOAT_NO_MARSHAL	(1<<4)
#define SERIALIZE_PAIR_NO_MARSHAL	(1<<5)
#define SERIALIZE_LIST_NO_MARSHAL	(1<<6)
#define SERIALIZE_ALL_NO_MARSHAL	(0xFFFF)
#define SERIALIZE_ANY_NO_MARSHAL	(0xFFFF)
#define SERIALIZE_ALL_AS_MARSHALED	(0x0)
			
typedef guint serialize_as_t;

/**
 * Takes a GConfValue and creates a single string representation of it
 *
 * See also ___external_serialize_value_from_string()
 *
 * Caller must free return value.
 *
 */
char *
___external_serialize_value_to_string (GConfValue * val);

/**
 * Creates a GConfValue from a single string representation of it what was
 * obtained using ___external_serialize_value_to_string()
 *
 * See also ___external_serialize_value_to_string()
 *
 * Caller must free return value.
 *
 */
GConfValue *
___external_serialize_value_from_string (const char *srz);

/**
 * If we have vtables (interfaces) why not version them too.
 */
#define SERIALIZE_FROM_GCONF_VALUE_VTABLE_VERSION_1 	1
typedef struct serialize_from_gconf_value_vtable_t serialize_from_gconf_value_vtable;

/**
 * Note that version and marshaled_f() are mandatory, others can be NULL
 */
struct serialize_from_gconf_value_vtable_t {
	guint version;	
	void (*type_and_val_f)( GError** err, const gchar* key, const gchar* type, gchar* value, gpointer user_data );
	void (*pair_f)( GError** err, const gchar* key, const gchar* type, const gchar* car_type, const gchar* cdr_type, gchar* car_value, gchar* cdr_value, gpointer user_data );
	void (*list_f)( GError** err, const gchar* key, const gchar* type, char** values, gpointer user_data );
	void (*marshaled_f)( GError** err, const gchar* key, char* value, gpointer user_data );
};


/**
 * demux the save operation, if the user wants some values to be saved as value & type
 * then demux that to the correct save function, otherwise, just marshal the data to
 * a single string and call a generic function to save that.
 *
 * Note that the default operation is the save the marshaled single string unless
 * the user has requested explicitly that values should be saved in a more
 * readable format.
 *
 * err Error place
 * key key to save
 * val Value to save
 * sat Special feild that contains user prefs for what way to save each value type.
 * 
 */
void
serialize_from_gconf_value( 
	GError** err, 
	const gchar* key, 
	GConfValue* value, 
	serialize_as_t sat,
	gpointer user_data,
	serialize_from_gconf_value_vtable* vtab);



/**
 * If we have vtables (interfaces) why not version them too.
 */
#define DESERIALIZE_TO_GCONF_VALUE_VTABLE_VERSION_1 	1
typedef struct deserialize_to_gconf_value_vtable_t deserialize_to_gconf_value_vtable;

/**
 * Note that version and marshaled_f() are mandatory, others can be NULL
 */
struct deserialize_to_gconf_value_vtable_t {
	guint version;	
	gchar* (*read_type_f)( GError** err, const gchar* key, gpointer user_data );
	GConfValue* (*val_f)( GError** err, const gchar* key, gchar* type, gpointer user_data );
	GConfValue* (*pair_f)( GError** err, const gchar* key, gchar* type, gpointer user_data );
	GConfValue* (*list_f)( GError** err, const gchar* key, gchar* type, gpointer user_data );
	gchar* (*marshaled_f)( GError** err, const gchar* key, gpointer user_data );
};

/**
 * demux the load operation, if the user wants some values to be saved/loaded as value & type
 * then demux that to the correct load function, otherwise, just marshal the data to
 * a single string and call a generic function to load that.
 *
 * Note that the default operation is the load the marshaled single string unless
 * the user has requested explicitly that values should be saved in a more
 * readable format.
 *
 * Note that there is read_type_f here and there was no save_type_f in  serialize_from_gconf_value()
 * This exists both out of need and due to the fact that a load at this point is usually out of
 * RAM. If there were a save_type_f then there may need to be two lots of network IO in 
 * serialize_from_gconf_value().
 *
 * Note that the return value from read_type_f() and marshaled_f() are freed by this function.
 * 
 *
 * err Error place
 * key key to save
 * sat Special feild that contains user prefs for what way to save each value type.
 * 
 */
GConfValue*
deserialize_to_gconf_value( 
	GError** err, 
	const gchar* key, 
	serialize_as_t sat,
	gpointer user_data,
	deserialize_to_gconf_value_vtable* vtab);

#endif /* ifndef INCLUDED_SERIALIZE_H */
